#ifndef MN_ASSET_MANAGER_
#define MN_ASSET_MANAGER_

		//TODO: add time stamps to log enteries: http://www.codersource.net/cpp_date_time.html

#include <stdio.h>
#include <string>
#include <list>
#include <fstream>
#include <iostream>

#define ASSETLOG	//Comment this out to disable logging of asset loading/unloading.



template< typename T >
struct AssetHolder {	//Bundles the asset with an ID and a file name so the the asset manager can track it
        T *assetPtr;
        std::string filename;
        unsigned int ID;	//Optional use, for superclasses.
};

/*This class is ment to be extended to implement the loading and tracking of assets to make sure they are not loaded twice, and to 
pass on pointers to loaded assets.*/
template< typename T >
class AssetManager
{
public:
        /*Creates an asset manager and if logging is compiled in it will log loads and unloads to "logFile".*/
        AssetManager(const char *logFile);
        /*Deletes the asset manager and all assets it was tracking.*/
        virtual ~AssetManager();

        /*Returns the AssetHolder struct with "filename" as a file name. If it does not exist, it returns NULL.*/
        T* checkLoaded(char *filename);

        /*Returns a pointer to a structure created from a file, making sure that the file was not loaded twice. Also assigns the id of the returned object to uid.*/
        T* Aquire(char *filename, unsigned int &uid, void *userData = NULL);

        /*Returns a pointer to a structure created from a file, making sure that the file was not loaded twice.*/
        T* Aquire(char *filename, void *userData = NULL);

        /*Returns the pointer to the type if it has been loaded, else returns NULL*/
        T* isIDLoaded(unsigned int ID);

        /*Deletes an object, by its ID, and returns true if the ID was in the list, false otherwise.*/
        bool DeleteByID(unsigned int ID);
        /*Deletes an object, by its memory address, and returns true if the object was found, false otherwise.*/
        bool DeleteByPtr(T* ptr);
        /*Deletes an object by its file name. If there is no such object the function returns false.*/
        bool DeleteByFilename(char *filename);
protected:
        /*Internal counter for creating userIDs, incrementation/creation should be handled in Load.*/
        unsigned int uIDCount;
        /*This is up the the superclass to implement, since this is general asset tracking class. Returns true if the asset was
        sucessfully loaded and was added to the asset tracking list.*/
        virtual T* Load(char *filename, unsigned int &userID, void *userData = NULL) = 0;

        #ifdef ASSETLOG
        std::string logFileName;
        std::ofstream logFile;
        #endif

private:
        /*A list of pointers to assets this class is tracking.*/
        std::list< AssetHolder<T> > assetList;
        /*If logging is enabled, the log file name to log to.*/
};


/////////////IMPLEMENTATION////////////


        //Constructor
template< typename T > 
AssetManager<T>::AssetManager(const char *logFile)
{
        uIDCount = 0;
        #ifdef ASSETLOG
                logFileName = logFile;
                AssetManager::logFile.open(logFile);
        #endif
}

template< typename T > 
T* AssetManager<T>::isIDLoaded(unsigned int ID)
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                if(iter->ID==ID)
                        return iter->assetPtr;
                iter++;
        }
        return NULL;
}

        //Destructor
template< typename T > 
AssetManager<T>::~AssetManager()
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                delete iter->assetPtr;
                #ifdef ASSETLOG
                        logFile<<"II - Unloaded "<<iter->filename.c_str()<<"\n";
                        logFile.flush();
                #endif
                iter++;
        }
        #ifdef ASSETLOG
        logFile.close();
        #endif
}

        //The checkLoaded function
template< typename T > 
T* AssetManager<T>::checkLoaded(char *filename)
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                if(strcmp(iter->filename.c_str(), filename)==0)
                        return iter->assetPtr;
                iter++;
        }
        return NULL;
}

template< typename T > 
T* AssetManager<T>::Aquire(char *filename, unsigned int &uid, void *userData)
{
        T* check = checkLoaded(filename);
        if(check==NULL)
        {
                unsigned int userID;
                T *loadedPtr = Load(filename, userID, userData);
                if(loadedPtr!=NULL)
                {
                        AssetHolder<T> newAssetHolder;
                        newAssetHolder.assetPtr = loadedPtr;
                        newAssetHolder.filename = filename;
                        newAssetHolder.ID = userID;
                        uid = userID;
                        assetList.push_back(newAssetHolder);
                        #ifdef ASSETLOG
                        logFile<<"II - Loaded "<<filename<<"\n";
                        logFile.flush();
                        #endif
                        return loadedPtr;
                } else
                {
                        #ifdef ASSETLOG
                        logFile<<"EE - Failed to load "<<filename<<"\n";
                        logFile.flush();
                        #endif
                        return NULL;
                }
        } else
                return check;
}

template< typename T > 
T* AssetManager<T>::Aquire(char *filename, void *userData)
{
        T* check = checkLoaded(filename);
        if(check==NULL)
        {
                unsigned int userID;
                T *loadedPtr = Load(filename, userID, userData);
                if(loadedPtr!=NULL)
                {
                        AssetHolder<T> newAssetHolder;
                        newAssetHolder.assetPtr = loadedPtr;
                        newAssetHolder.filename = filename;
                        newAssetHolder.ID = userID;
                        assetList.push_back(newAssetHolder);
                        #ifdef ASSETLOG
                        logFile<<"II - Loaded "<<filename<<"\n";
                        logFile.flush();
                        #endif
                        return loadedPtr;
                } else
                {
                        #ifdef ASSETLOG
                        logFile<<"EE - Failed to load "<<filename<<"\n";
                        logFile.flush();
                        #endif
                        return NULL;
                }
        } else
                return check;
}

template< typename T > 
bool AssetManager<T>::DeleteByID(unsigned int ID)
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                if(iter->ID == ID)
                {
                        #ifdef ASSETLOG
                        logFile<<"II - Unloaded "<<iter->filename<<" by ID. ID="<<ID<<std::endl;
                        logFile.flush();
                        #endif
                        delete iter->assetPtr;
                        assetList.erase(iter);
                        return true;
                }
                iter++;
        }
        #ifdef ASSETLOG
        logFile<<"WW - Tried to unload an object by ID, but the object doesn't exist. ID="<<ID<<std::endl;
        logFile.flush();
        #endif
        return false;
}

template< typename T > 
bool AssetManager<T>::DeleteByPtr(T* ptr)
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                if(iter->assetPtr == ptr)
                {
                        #ifdef ASSETLOG
                        logFile<<"II - Unloaded "<<iter->filename<<" by pointer."<<std::endl;
                        logFile.flush();
                        #endif
                        delete iter->assetPtr;
                        assetList.erase(iter);
                        return true;
                }
                iter++;
        }
        #ifdef ASSETLOG
        logFile<<"WW - Tried to unload an object by pointer, but the object doesn't exist."<<std::endl;
        logFile.flush();
        #endif
        return false;
}

template< typename T > 
bool AssetManager<T>::DeleteByFilename(char *filename)
{
        typename std::list< AssetHolder<T> >::iterator iter = assetList.begin();
        while(iter!=assetList.end())
        {
                if(strcmp(iter->filename, filename)==0)
                {
                        #ifdef ASSETLOG
                        logFile<<"II - Unloaded "<<iter->filename<<" by filename."<<std::endl;
                        logFile.flush();
                        #endif
                        delete iter->assetPtr;
                        assetList.erase(iter);
                        return true;
                }
                iter++;
        }
        #ifdef ASSETLOG
        logFile<<"WW - Tried to unload an object by filename, but the object doesn't exist. The filename was '"<<filename<<"'"<<std::endl;
        logFile.flush();
        #endif
        return false;
}


#endif

